@@ -21,7 +21,6 @@
 
 from queue import Queue
 import socket
-import zmq
 import threading
 from threading import Thread, Event
 import _thread
@@ -32,16 +31,17 @@ import os
 import inspect
 from inspect import signature
 
-from IPFS_API import *
-import IPFS_API
-IPFS_API.Start()
-zmq_context = zmq.Context()
+try:
+    import IPFS_API
+except:
+    import IPFS_API_Remote_Client as IPFS_API
+# IPFS_API.Start()
 
 
 # -------------- Settings ---------------------------------------------------------------------------------------------------
 print_log = False  # whether or not to print debug in output terminal
 print_log_connections = False
-print_log_transmissions = False
+print_log_transmissions = True
 print_log_conversations = True
 print_log_files = True
 
@@ -73,7 +73,8 @@ def TransmitData(data: bytes, peerID: str, req_lis_name: str, timeout_sec: int =
     Transmits the input data (a bytearray of any length) to the computer with the specified IPFS peer ID.
 
     Usage:
-        success = TransmitData("data to transmit".encode("utf-8"), "Qm123456789", "applicationNo2")    # transmits "data to transmit" to the computer with the Peer ID "Qm123456789", for the IPFS_DataTransmission listener called "applicationNo2" at a buffersize of 1024 bytes
+        # transmits "data to transmit" to the computer with the Peer ID "Qm123456789", for the IPFS_DataTransmission listener called "applicationNo2" at a buffersize of 1024 bytes
+        success = TransmitData("data to transmit".encode("utf-8"), "Qm123456789", "applicationNo2")
 
     Parameters:
         bytearray data: the data to be transmitted to the receiver
@@ -84,66 +85,57 @@ def TransmitData(data: bytes, peerID: str, req_lis_name: str, timeout_sec: int =
     Returns:
         bool success: whether or not the transmission succeeded
     """
+    if peerID == IPFS_API.MyID():
+        return
+
     def SendTransmissionRequest():
         # sending transmission request, telling the receiver our code for the transmission, our listening port on which they should send confirmation buffers, and the buffer size to use
-        # sock.connect(("127.0.0.1", port))
         data = AddIntegrityByteToBuffer(IPFS_API.MyID().encode(
-        ) + bytearray([255]) + ToB255No0s(our_port))
+        ))
         tries = 0
-        poller = zmq.Poller()
         while max_retries == -1 or tries < max_retries:
-            sock = CreateSendingConnection(peerID, req_lis_name)
-            success = sock.send(data)
             if print_log_transmissions:
-                print(str(our_port)
-                      + ": Sent transmission request to " + str(req_lis_name))
+                print("Sending transmission request to " + str(req_lis_name))
+            sock = CreateSendingConnection(peerID, req_lis_name)
+            sock.sendall(data)
+            if print_log_transmissions:
+                print("Sent transmission request to " + str(req_lis_name))
 
-            poller.register(sock, zmq.POLLIN)
-            evts = poller.poll(timeout_sec*1000)
-
-            poller.unregister(sock)
+            reply = sock.recv(def_buffer_size)
             sock.close()
             del sock
             CloseSendingConnection(peerID, req_lis_name)
-            if len(evts) > 0:
+            if reply:
                 if print_log_transmissions:
-                    print(str(our_port)
-                          + ": Transmission request to " + str(req_lis_name) + "was received.")
-                return True  # signal success
+                    print("Transmission request to " + str(req_lis_name) + "was received.")
+                return reply[30:].decode()  # signal success
             else:
                 if print_log_transmissions:
-                    print(str(our_port)
-                          + ": Transmission request send " + str(req_lis_name) + "timeout_sec reached.")
+                    print("Transmission request send " + str(req_lis_name) + "timeout_sec reached.")
             tries += 1
         CloseSendingConnection(peerID, req_lis_name)
         return False    # signal Failure
 
-    sock = zmq_context.socket(zmq.REP)
-    our_port = sock.bind_to_random_port("tcp://*")
-    CreateListeningConnection(str(our_port), our_port)
+    their_trsm_port = SendTransmissionRequest()
+    if their_trsm_port:
+        sock = CreateSendingConnection(peerID, their_trsm_port)
+        sock.sendall(data)  # transmit Data
+        if print_log_transmissions:
+            print("Sent Transmission Data", data)
+        response = sock.recv(def_buffer_size)
+        if response and response == b"Finished!":
+            # conn.close()
+            sock.close()
+            CloseSendingConnection(peerID, their_trsm_port)
+            if print_log_transmissions:
+                print(": Finished transmission.")
+            return True  # signal success
+        else:
+            if print_log_transmissions:
+                print("Received unrecognised response:", response)
 
-    poller = zmq.Poller()
-    poller.register(sock, zmq.POLLIN)
-    if SendTransmissionRequest():
-        evts = poller.poll(timeout_sec*1000)    # wait for receiver to signal readiness
-        if len(evts) > 0 and sock.recv() == b"start transmission":
-            poller = zmq.Poller()
-            poller.register(sock, zmq.POLLIN)
-            sock.send(data)  # transmit Data
-            evts = poller.poll(timeout_sec*1000)
-
-            if len(evts) > 0 and sock.recv() == b'Finished!':
-                sock.close()
-
-                CloseListeningConnection(str(our_port), our_port)
-
-                if print_log_transmissions:
-                    print(str(our_port) + ": Finished transmission.")
-
-                return True  # signal success
-
-    sock.close()
-    CloseListeningConnection(str(our_port), our_port)
+    # sock.close()
+    # CloseSendingConnection(peerID, their_trsm_port)
     return False    # signal Failure
 
 
@@ -218,18 +210,21 @@ class TransmissionListener:
                         self.listener_name + ": Received a buffer with a non-matching integrity buffer")
                 return
 
-            index = data.index(bytearray([255]))
-            peerID = data[0:index].decode()
-            data = data[index + 1:]
-            sender_port = FromB255No0s(data)
+            peerID = data.decode()
 
             if print_log_transmissions:
                 print(
                     self.listener_name + ": Received transmission request.")
+            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+            sock.bind(("127.0.0.1", 0))
+            our_port = sock.getsockname()[1]
+            CreateListeningConnection(str(our_port), our_port)
+            sock.listen()
+
             listener = Thread(target=self.ReceiveTransmission, args=(
-                peerID, sender_port, self.eventhandler))
+                peerID, sock, our_port, self.eventhandler))
             listener.start()
-            return listener
+            return our_port
 
         except Exception as e:
             print("")
@@ -242,56 +237,60 @@ class TransmissionListener:
             print(e)
             print(self.listener_name + ": Could not decode transmission request.")
 
-    def ReceiveTransmission(self, peerID, sender_port, eventhandler):
-        sock = CreateSendingConnection(peerID, str(sender_port))
-
+    def ReceiveTransmission(self, peerID, sock, our_port, eventhandler):
+        #
+        # sock = CreateSendingConnection(peerID, str(sender_port))
+        #
+        # if print_log_transmissions:
+        #     print("Ready to receive transmission.")
+        #
+        # sock.sendall(b"start transmission")
         if print_log_transmissions:
-            print("Ready to receive transmission.")
-
-        poller = zmq.Poller()
-        poller.register(sock, zmq.POLLIN)
-
-        sock.send(b"start transmission")
-        evts = poller.poll(transmission_receive_timeout_sec*1000)
-        if len(evts) == 0:
-            CloseSendingConnection(peerID, str(sender_port))
-            if print_log:
-                print("Transmission reception failed.")
-            return
-        data = sock.recv()
-        sock.send("Finished!".encode())
-        sock.close()
+            print("waiting to receive actual transmission")
+        conn, addr = sock.accept()
+        if print_log_transmissions:
+            print("received connection response fro actual transmission")
+        data = recv_timeout(conn)
+        conn.send("Finished!".encode())
+        # conn.close()
         _thread.start_new_thread(eventhandler, (data, peerID))
-        CloseSendingConnection(peerID, str(sender_port))
+        CloseListeningConnection(str(our_port), our_port)
+        sock.close()
 
     def Listen(self):
-        # context = zmq.Context()
-        self.socket = zmq_context.socket(zmq.REP)
-        self.port = self.socket.bind_to_random_port("tcp://*")
+        if print_log_transmissions:
+            print("Creating Listener")
+        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        self.socket.bind(("127.0.0.1", 0))
+        self.port = self.socket.getsockname()[1]
         CreateListeningConnection(self.listener_name, self.port)
+
         if print_log_transmissions:
             print(self.listener_name
                   + ": Listening for transmission requests as " + self.listener_name)
-        while self.socket:
-            data = self.socket.recv()
+        self.socket.listen()
+        while True:
+            conn, addr = self.socket.accept()
+            data = recv_timeout(conn)
             if self.terminate:
-                self.socket.send(b"Righto.")
+                # conn.sendall(b"Righto.")
                 return
-            if self.ReceiveTransmissionRequests(data):
-                self.socket.send(b"Transmission request accepted.")
+            port = self.ReceiveTransmissionRequests(data)
+            if port:
+                conn.send(f"Transmission request accepted.{port}".encode())
             else:
-                self.socket.send(b"Transmission request not accepted.")
+                conn.send(b"Transmission request not accepted.")
 
     def Terminate(self):
         # self.socket.unbind(self.port)
         self.terminate = True
         CloseListeningConnection(self.listener_name, self.port)
-        socket = zmq_context.socket(zmq.REQ)
-        socket.connect(f"tcp://localhost:{self.port}")
-        socket.send("close".encode())
-        socket.recv()
-        socket.close()
-        del socket
+        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        sock.connect(("127.0.0.1", self.port))
+        sock.sendall("close".encode())
+        # recv_timeout(sock)
+        sock.close()
+        del sock
         self.socket.close()
         del self.socket
         del self.listener
@@ -311,7 +310,7 @@ def StartConversation(conversation_name, peerID, others_req_listener, eventhandl
         conversation = StartConversation("test", "QmHash", "conv listener", OnReceive)
         conversation.Say("Hello there!".encode())
 
-        ## The other peer must have a conversation listener named "conv listener" running, e.g. via:
+        # The other peer must have a conversation listener named "conv listener" running, e.g. via:
         # forwards communication from any conversation to the OnReceive function
         ListenForConversations("conv listener", OnReceive)
 
@@ -325,18 +324,11 @@ def StartConversation(conversation_name, peerID, others_req_listener, eventhandl
     return conv
 
 
-def StartConversationAwait(conversation_name, peerID, others_req_listener, eventhandler=None):
-    print("IPFS_API: WARNING: This function is deprecated.Please use StartConversation() instead.\n" +
-          "The reason for this is the introduction of a proper failure handling system.\n")
-    return StartConversation(conversation_name, peerID,
-                             others_req_listener, eventhandler)
-
-
 def ListenForConversations(conv_name, eventhandler):
     """
     Listen to incoming conversation requests.
     Whenever a new conversation request is received, the specified eventhandler
-    is called which must then decide whether or not to join the conversation, 
+    is called which must then decide whether or not to join the conversation,
     and then act upon that decision.
 
     Usage Example:
@@ -367,7 +359,7 @@ class Conversation:
     Communication object which allows 2 peers to repetatively make
     data transmissions to each other asynchronously and bidirectionally.
     Usage example:
-        ## Conversation Initiator:
+        # Conversation Initiator:
         def OnReceive(conversation, data):
             print(data.decode("utf-8"))
 
@@ -375,7 +367,7 @@ class Conversation:
         conversation.Start("test", "QmHash", "conv listener", OnReceive)
         conversation.Say("Hello there!".encode())
 
-        ## The other peer must have a conversation listener named "conv listener" running, e.g. via:
+        # The other peer must have a conversation listener named "conv listener" running, e.g. via:
         # forwards communication from any conversation to the OnReceive function
         ListenForConversations("conv listener", OnReceive)
 
@@ -391,17 +383,21 @@ class Conversation:
                 (can be used as an alternative to the event handler which is
                 optionally specified in the Start() or Join() methods)
     """
-    conversation_started = False
-    # started = Event()
-    # last_message = None
-    # message_received = Event()
+    # file_progress_callback = None
     eventhandler = None
-    message_queue = Queue()
+    file_eventhandler = None
+    file_progress_callback = None
 
     def __init__(self):
         self.started = Event()
+        self.conversation_started = False
+        self.eventhandler = None
+        self.file_eventhandler = None
+        self.file_progress_callback = None
+        self.message_queue = Queue()
+        self.file_queue = Queue()
 
-    def Start(self, conversation_name, peerID, others_req_listener, eventhandler=None, timeout_sec=transmission_send_timeout_sec, max_retries=transmission_request_max_retries):
+    def Start(self, conversation_name, peerID, others_req_listener, eventhandler=None, file_eventhandler=None, file_progress_callback=None, timeout_sec=transmission_send_timeout_sec, max_retries=transmission_request_max_retries):
         """Starts a conversation object with which peers can send data transmissions to each other.
         Code blocks until the other peer joins the conversation or timeout is reached
         Usage:
@@ -424,15 +420,20 @@ class Conversation:
         Returns:
             bool success: whether or not the conversation with the peer was successfully started
         """
+        if peerID == IPFS_API.MyID():
+            return
         if(print_log_conversations):
             print(conversation_name + ": Starting conversation")
         self.conversation_name = conversation_name
         self.eventhandler = eventhandler
-
+        self.file_eventhandler = file_eventhandler
+        self.file_progress_callback = file_progress_callback
         self.peerID = peerID
         if print_log_conversations:
             print(conversation_name + ": sending conversation request")
         self.listener = ListenForTransmissions(conversation_name, self.Hear)
+        self.file_receiver = ListenForFileTransmissions(
+            f"{conversation_name}:files", self.FileReceived, self.file_progress_callback)
         # self.listener = ListenForTransmissions(conversation_name, self.hear_eventhandler)
         data = bytearray("I want to start a conversation".encode(
             'utf-8')) + bytearray([255]) + bytearray(conversation_name.encode('utf-8'))
@@ -444,12 +445,7 @@ class Conversation:
         else:
             return False    # signal Failure
 
-    def StartAwait(self, conversation_name, peerID, others_req_listener, eventhandler=None):
-        print("IPFS_API: WARNING: This function is deprecated.Please use StartConversation() instead.\n" +
-              "The reason for this is the introduction of a proper failure handling system.\n")
-        return self.Start(self, conversation_name, peerID, others_req_listener, eventhandler=None)
-
-    def Join(self, conversation_name, peerID, others_trsm_listener, eventhandler=None):
+    def Join(self, conversation_name, peerID, others_trsm_listener, eventhandler=None, file_eventhandler=None, file_progress_callback=None):
         """
         Joins a conversation which another peer started, given their peer ID
         and conversation's transmission-listener's name.
@@ -462,7 +458,11 @@ class Conversation:
             print(conversation_name + ": Joining conversation "
                   + others_trsm_listener)
         self.eventhandler = eventhandler
+        self.file_eventhandler = file_eventhandler
+        self.file_progress_callback = file_progress_callback
         self.listener = ListenForTransmissions(conversation_name, self.Hear)
+        self.file_receiver = ListenForFileTransmissions(
+            f"{conversation_name}:files", self.FileReceived, self.file_progress_callback)
         self.others_trsm_listener = others_trsm_listener
         self.peerID = peerID
         data = bytearray("I'm listening".encode(
@@ -531,10 +531,32 @@ class Conversation:
             return data
         else:
             if print_log_conversations:
-                print("Conv.Listen: received", self.last_message, "restarting Event Wait")
+                print("Conv.Listen: received nothing restarting Event Wait")
             self.Listen()
             return
 
+    def FileReceived(self, peer, filepath, metadata):
+        self.file_queue.put(filepath)
+        print(filepath)
+        if self.file_eventhandler:
+            _thread.start_new_thread(self.file_eventhandler, (self, filepath, metadata))
+
+    def ListenForFile(self, timeout=None):
+        if not timeout:
+            data = self.file_queue.get()
+        else:
+            try:
+                data = self.file_queue.get(timeout=timeout)
+            except:  # timeout reached
+                return None
+        if data:
+            return data
+        else:
+            if print_log_conversations:
+                print("Conv.FileListen: received nothign restarting Event Wait")
+            self.ListenForFile(timeout)
+            return
+
     def Say(self, data, timeout_sec=transmission_send_timeout_sec, max_retries=transmission_request_max_retries):
         """
         Transmits the input data (a bytearray of any length) to the other computer in this conversation.
@@ -557,6 +579,28 @@ class Conversation:
             time.sleep(0.01)
         return TransmitData(data, self.peerID, self.others_trsm_listener, timeout_sec, max_retries)
 
+    def TransmitFile(self, filepath, metadata=bytearray(), progress_handler=file_progress_callback, block_size=def_block_size):
+        """
+        Transmits the input data (a bytearray of any length) to the other computer in this conversation.
+
+        Usage:
+            success = conv.Say("data to transmit".encode("utf-8"), "Qm123456789", "applicationNo2")    # transmits "data to transmit" to the computer with the Peer ID "Qm123456789", for the IPFS_DataTransmission listener called "applicationNo2" at a buffersize of 1024 bytes
+
+        Parameters:
+            bytearray data: the data to be transmitted to the receiver
+            string peerID: the IPFS peer ID of [the recipient computer to send the data to]
+            string listener_name: the name of the IPFS-Data-Transmission-Listener instance running on the recipient computer to send the data to (allows distinguishing multiple IPFS-Data-Transmission-Listeners running on the same computer for different applications)
+            timeout_sec: connection attempt timeout, multiplied with the maximum number of retries will result in the total time required for a failed attempt
+            max_retries: how often the transmission should be reattempted when the timeout is reached
+        Returns:
+            bool success: whether or not the transmission succeeded
+        """
+        while not self.conversation_started:
+            if print_log:
+                print("Wanted to say something but conversation was not yet started")
+            time.sleep(0.01)
+        return TransmitFile(filepath, self.peerID, f"{self.others_trsm_listener}:files", metadata, progress_handler, block_size)
+
     def Close(self):
         self.listener.Terminate()
 
@@ -674,7 +718,8 @@ def ListenForFileTransmissions(listener_name, eventhandler, progress_handler=Non
 class FileTransmitter:
     """Transmits the given file to the specified peer
     Usage:
-        transmitter = TransmitFile("text.txt", "QMHash", "filelistener", "testmeadata".encode())
+        file_transmitter = FileTransmitter("text.txt", "QMHash", "filelistener", "testmeadata".encode())
+        file_transmitter.Start()
     Paramaters:
         string filepath: the path of the file to transmit
         string peerID: the IPFS ID of the computer to send the file to
@@ -689,6 +734,8 @@ class FileTransmitter:
     status = "not started"  # "transitting" "finished" "aborted"
 
     def __init__(self, filepath, peerID, others_req_listener, metadata=bytearray(), progress_handler=None, block_size=def_block_size):
+        if peerID == IPFS_API.MyID():
+            return
         self.filesize = os.path.getsize(filepath)
         self.filename = os.path.basename(filepath)
         self.filepath = filepath
@@ -698,10 +745,6 @@ class FileTransmitter:
         self.others_req_listener = others_req_listener
         self.progress_handler = progress_handler
 
-        print("IPFS_API: WARNING: In this new version the FileTransmitter no longer starts automatically.\n" +
-              "Call the Start() function to start the transmission." +
-              "The reason for this is the introduction of a proper failure handling system.\n")
-
     def Start(self):
         """
         Returns:
@@ -717,8 +760,7 @@ class FileTransmitter:
             if print_log_files:
                 print("FileTransmission: " + self.filename
                       + ": Sent transmission request")
-            if self.progress_handler:
-                _thread.start_new_thread(self.progress_handler, (0,))
+            self.CallProgressCallBack(0)
             return True
         else:
             return False
@@ -736,8 +778,7 @@ class FileTransmitter:
                     blocksize = self.block_size
                 data = reader.read(blocksize)
                 position += len(data)
-                if self.progress_handler:
-                    _thread.start_new_thread(self.progress_handler, (position/self.filesize,))
+                self.CallProgressCallBack(position/self.filesize,)
 
                 if print_log_files:
                     print("FileTransmission: " + self.filename
@@ -749,6 +790,16 @@ class FileTransmitter:
         self.status = "finished"
         self.conversation.Close()
 
+    def CallProgressCallBack(self, progress):
+        if self.progress_handler:
+            if len(signature(self.progress_handler).parameters) == 1:
+                _thread.start_new_thread(self.progress_handler, (progress,))
+            elif len(signature(self.progress_handler).parameters) == 2:
+                _thread.start_new_thread(self.progress_handler, (self.filename, progress))
+            elif len(signature(self.progress_handler).parameters) == 3:
+                _thread.start_new_thread(self.progress_handler,
+                                         (self.filename, self.filesize, progress))
+
     def Hear(self, conv, data):
         if print_log_files:
             print("FileTransmission: " + self.filename
@@ -841,12 +892,13 @@ class FileTransmissionReceiver:
                   + ": Transmission finished.")
         self.status = "finished"
         self.conv.Close()
-        if signature(self.eventhandler).parameters.get("metadata") != None:
-            self.eventhandler(self.conv.peerID, os.path.join(
-                self.dir, self.filename), self.metadata)
-        else:
-            self.eventhandler(self.conv.peerID, os.path.join(
+        if self.eventhandler:
+            filepath = os.path.abspath(os.path.join(
                 self.dir, self.filename))
+            if signature(self.eventhandler).parameters["metadata"]:
+                self.eventhandler(self.conv.peerID, filepath, self.metadata)
+            else:
+                self.eventhandler(self.conv.peerID, filepath)
 
 
 # ----------IPFS Technicalitites-------------------------------------------
@@ -858,27 +910,27 @@ def CreateSendingConnection(peerID: str, protocol: str, port=None):
     IPFS_API.ClosePortForwarding(
         targetaddress="/p2p/" + peerID, protocol="/x/" + protocol)
 
-    sock = zmq_context.socket(zmq.REQ)
-    # socket.CONNECT_TIMEOUT = 1
+    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
     if port == None:
         for prt in sending_ports:
             try:
-                ForwardFromPortToPeer(protocol, prt, peerID)
-                sock.connect(f"tcp://localhost:{prt}")
+                IPFS_API.ForwardFromPortToPeer(protocol, prt, peerID)
+                sock.connect(("127.0.0.1", prt))
+
                 return sock
             except:
                 pass
 
         print("failed to find free port for sending connection")
     else:
-        ForwardFromPortToPeer(protocol, port, peerID)
-        sock.connect(f"tcp://localhost:{prt}")
+        IPFS_API.ForwardFromPortToPeer(protocol, port, peerID)
+        sock.connect(("127.0.0.1", prt))
         return sock
 
 
 def CreateListeningConnectionTCP(protocol, port):
     try:
-        ListenOnPort(protocol, port)
+        IPFS_API.ListenOnPort(protocol, port)
         if print_log_connections:
             print(f"listening as \"{protocol}\" on {port}")
     except Exception as e:
@@ -886,7 +938,7 @@ def CreateListeningConnectionTCP(protocol, port):
             listenaddress=f"/ip4/127.0.0.1/tcp/{port}")
         IPFS_API.ClosePortForwarding(protocol=f"/x/{protocol}")
         try:
-            ListenOnPort(protocol, port)
+            IPFS_API.ListenOnPort(protocol, port)
         except:
             print(f"listening as \"{protocol}\" on {port}")
             print("")
@@ -905,7 +957,7 @@ def CreateListeningConnectionTCP(protocol, port):
 
 def CreateListeningConnection(protocol, port):
     try:
-        ListenOnPort(protocol, port)
+        IPFS_API.ListenOnPort(protocol, port)
         if print_log_connections:
             print(f"listening fas \"{protocol}\" on {port}")
     except Exception as e:
@@ -914,7 +966,7 @@ def CreateListeningConnection(protocol, port):
         IPFS_API.ClosePortForwarding(protocol=f"/x/{protocol}")
         try:
             time.sleep(0.1)
-            ListenOnPort(protocol, port)
+            IPFS_API.ListenOnPort(protocol, port)
             if print_log_connections:
                 print(f"listening fas \"{protocol}\" on {port}")
         except:
@@ -1414,3 +1466,39 @@ def SplitBy255(bytes):
         pos += 1
     result.append(bytearray(collected))
     return result
+
+
+def recv_timeout(the_socket, timeout=2):
+    # make socket non blocking
+    the_socket.setblocking(0)
+
+    # total data partwise in an array
+    total_data = bytearray()
+    data = bytearray()
+
+    # beginning time
+    begin = time.time()
+    while 1:
+        # if you got some data, then break after timeout
+        if len(total_data) > 0 and time.time()-begin > timeout:
+            break
+
+        # if you got no data at all, wait a little longer, twice the timeout
+        elif time.time()-begin > timeout*2:
+            break
+
+        # recv something
+        try:
+            data = the_socket.recv(def_buffer_size)
+            if data:
+                total_data += data
+                # change the beginning time for measurement
+                begin = time.time()
+            # else:
+            #     # sleep for sometime to indicate a gap
+            #     time.sleep(0.1)
+        except:
+            pass
+
+    # print("RECEIVED", type(total_data), total_data)
+    return total_data
